/* $Id$ */

/* seek.c -- functions to get plaintext from media files --
 * Copyright (C) 2000 Synack Systems Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <png.h>
#include <errno.h>
#include <setjmp.h>
#include "common.h"
#include "seek.h"
#include "blowfish.h"
#include "setup.h"

int percentage = 10;
BLOWFISH_context bf;

byte *seek(u32 *datalen, byte *noise, u32 noiselen, u32 noiseskip, byte noiseunitsz, u32 randnum)
{
    u32 i, k, ratio, pos, next;
    HIDE_HEADER header;
    byte *p, bit;
    u32 hidespots[HIDE_HEADER_SZ][8];
    int h1, h2;
    byte *data=NULL;


    /* let's grab the header */
    memset(&header, 0, sizeof(header));
    p = (byte*)&header;
    for(i=0;i<sizeof(header);i++) {
        for(k=0;k<8;k++) {

            /* figure out where to store the next bit.. we have to advance pos
             * by at least 1 and make sure that pos doesn't land on a byte
             * that we shouldn't use
             */
            do {
                bfencrypt(&bf, &randnum, &randnum);
                next = randnum % noiselen;
            } while(!(next%noiseskip));

            hidespots[i][k] = pos = next;

            bit = noise[pos] & 0x01;
            p[i] |= (bit << k);

            // printf("reading bit %d at pos %d of %d\n", bit, pos, noiselen);
        }
    }

    *datalen = header.datalen = ntohl(header.datalen);

    /* we need to do some sanity checking now.. */
    if(header.datalen > noiselen/percentage) {
        fprintf(stderr, "datalen (%d) appears to be invalid; giving up..\n", header.datalen);
        goto cleanup;
    }

    if((header.ver != 0) || (header.ver != 1)) {
        fprintf(stderr, "unknown refugee header version (%d).  giving up..\n", header.ver);
        goto cleanup;
    }

    printf("datalen is %u; percentage is %d.\n", header.datalen, header.percent);

    /* see hide.c for an explanation of how we come up with ratio */
    ratio = noiselen / ((header.datalen+sizeof(header))*8);
    if(header.ver) ratio *= 1.7;

    printf("ratio: 1:%d\n", ratio);

    pos = 0;
    data = malloc(header.datalen);
    memset(data, 0, header.datalen);
    p = (byte*)data;

    for(i=0;i<header.datalen;i++) {
        for(k=0;k<8;k++) {

        tryagain:

            do {
                do {
                    bfencrypt(&bf, &randnum, &randnum);
                    next = randnum % ratio;
                } while(!next);
            } while(!((pos+next)%noiseskip));

            pos += next;

            /* make sure we aren't reusing a header bit */
            for(h1=0;h1<HIDE_HEADER_SZ;h1++)
                for(h2=0;h2<8;h2++)
                    if(hidespots[h1][h2] == pos)
                        goto tryagain;

            bit = noise[pos] & 0x01;
            p[i] |= (bit << k);

            //printf("reading bit %d at pos %d of %d\n", bit, pos, noiselen);

        }
    }

cleanup:
    memset(&header, 0, sizeof(header));
    memset(hidespots, 0, sizeof(hidespots));

    return data;
}



int png_seek(char *png_filename, char *out_filename, u32 key) {
    FILE *in=NULL, *out=NULL;
    int status=0, sig_read, i;
    byte header[4];
    png_structp png_ptr=NULL, png_write_ptr=NULL;
    png_infop info_ptr=NULL;
    png_uint_32 width, height;
    int depth, color_type, interlace_type;
    png_bytep *row_pointers=NULL;
    byte *image_data=NULL;
    u32 maxfit;
    u32 rowbytes, channels;
    u32 datalen;
    byte *data=NULL;

    if((in = fopen(png_filename, "rb")) == NULL) {
        fprintf(stderr, "couldn't open %s for reading.\n", png_filename);
        status=1; goto cleanup;
    }

    sig_read = fread(header, 1, 4, in);
    if(png_sig_cmp(header, 0, sig_read)) {
        fprintf(stderr, "%s is not a valid PNG file.\n", png_filename);
        status=1; goto cleanup;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(png_ptr == NULL) {
        fprintf(stderr, "unable to initialize PNG library (1)\n");
        status=1; goto cleanup;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(info_ptr == NULL) {
        fprintf(stderr, "unable to initialize PNG library (2)\n");
        status=1; goto cleanup;
    }

    if(setjmp(png_ptr->jmpbuf)) {
        fprintf(stderr, "an unrecoverable libpng error has occurred.\n");
        status=1; goto cleanup;
    }

    png_init_io(png_ptr, in);
    png_set_sig_bytes(png_ptr, sig_read);
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &depth, &color_type,
                 &interlace_type, NULL, NULL);

    /* maxfit is the max nunmber of bytes that can be stored in an image */
    /* we only work with the RGB components of the image and skip the alpha */
    /* only use percentage% of the pixels */
    /* takes 8 pixels to hold one byte */
    //maxfit = (width*height*3) / percentage / 8;
    //printf("%s is %ldx%ldx%d and can hold %d bytes of data.\n", fname, width,
    //       height, depth, maxfit);
    //if(maxfit < datalen+sizeof(HIDE_HEADER)) {
    //    fprintf(stderr, "Refugee can not hide your data.\n%s is not big enough!\n", fname);
    //    status=66; goto cleanup;
    //}

    //if(color_type == PNG_COLOR_TYPE_GRAY) {
    //    fprintf(stderr, "You need to use a color PNG file; %s is B&W.\n", fname);
    //    status=1; goto cleanup;
    //}

    /* we want to always work with 4 channels even though we're only going to
     * use 3 of them -- why?  to rest my weary brain!
     */
    if(color_type == PNG_COLOR_TYPE_PALETTE)
        png_set_expand(png_ptr);

    png_read_update_info(png_ptr, info_ptr);

    rowbytes = png_get_rowbytes(png_ptr, info_ptr);
    channels = png_get_channels(png_ptr, info_ptr);

    if((image_data = malloc(rowbytes*height)) == NULL) {
        fprintf(stderr, "malloc failed.\n");
        status=2; goto cleanup;
    }

    row_pointers = (png_bytepp)malloc(height*sizeof(png_bytep));
    if(!row_pointers) {
        fprintf(stderr, "malloc failed.\n");
        status=2; goto cleanup;
    }

    for(i=0;i<height;i++)
        row_pointers[i] = image_data + i*rowbytes;

    png_read_image(png_ptr, row_pointers);

    free(row_pointers);
    row_pointers = NULL;

    png_read_end(png_ptr, info_ptr);
    fclose(in); in=NULL;

    data=seek(&datalen, image_data, rowbytes*height, 4, depth, key);
    if(!data) {
        fprintf(stderr, "error recovering data from media file\n");
        status=1; goto cleanup;
    }

    if((out = fopen(out_filename, "wb")) == NULL) {
        fprintf(stderr, "couldn't open %s for writing.\n", out_filename);
        status=1; goto cleanup;
    }
    fwrite(data, datalen, 1, out);
    fclose(out);
    out=NULL;


cleanup:
    if(row_pointers) free(row_pointers);
    if(image_data) {
        memset(image_data, 0, rowbytes*height);
        free(image_data);
    }
    if(data) {
        memset(data, 0, datalen);
        free(data);
        datalen=0;
    }

    if(png_ptr)
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    if(png_write_ptr)
        png_destroy_write_struct(&png_write_ptr, NULL);

    memset(png_filename, 0, strlen(png_filename));
    memset(out_filename, 0, strlen(out_filename));

    datalen=0;
    key=0;

    if(in) fclose(in);
    if(out) fclose(out);

    return status;
}

int main(int argc, char **argv) {
    int status=0;
    int i;
    byte *bfkey=NULL;
    u32 keylen, *keyptr,randnum;


    if(argc != 3) {
        fprintf(stderr, "usage: %s png_filename data_filename\n", argv[0]);
        fprintf(stderr, "Refugee will find the data hidden in png_filename and save it to data_filename.\n");
        exit(1);
    }

    nocore();

    bfkey = getbfkey(&keylen);
    if(!bfkey) {
        fprintf(stderr, "Couldn't get a key.  Run key.\n");
        status=1; goto cleanup;
    }
    if(bf_setkey(&bf, bfkey, keylen) != 0) {
        fprintf(stderr, "error setting blowfish key.\n");
        status=1; goto cleanup;
    }

    randnum=0;
    keyptr = (u32*)bfkey;
    for(i=0;i<keylen/4;i++)
        randnum ^= ntohl(*(keyptr+i));

    status = png_seek(argv[1], argv[2], randnum);

cleanup:

    if(bfkey)
        memset(bfkey, 0, keylen);

    keylen = 0;

    memset(&bf, 0, sizeof(bf));

    return status;
}
